/*
 * Copyright (c) 2013, Altera Corporation.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  - Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */


/* Handling of special commands (anything that is not read/write/open/close)
 * that user may call.
 * See pcie_linux_driver_exports.h for explanations of each command. */


#include <linux/mm.h>
#include <linux/device.h>
#include <linux/sched.h>
#include <linux/aer.h>
//#include <linux/pagemap.h>
#include <linux/version.h>

#include "aclpci.h"

/* RedHat 5.5 doesn't define this */
#ifndef PCI_EXP_LNKSTA_NLW_SHIFT
#define PCI_EXP_LNKSTA_NLW_SHIFT 4
#endif


void get_pcie_params (struct aclpci_dev *aclpci);
void disable_aer_on_upstream_dev(struct aclpci_dev *aclpci);
void restore_aer_on_upstream_dev(struct aclpci_dev *aclpci);


/* Execute special command */
ssize_t aclpci_exec_cmd (struct aclpci_dev *aclpci, 
                         struct aclpci_cmd kcmd, 
                         size_t count) {
  ssize_t result = 0;

//  ACL_DEBUG("Entry with kcmd.command %d", kcmd.command);

  switch (kcmd.command) {
  case ACLPCI_CMD_SAVE_PCI_CONTROL_REGS: {
    ACL_DEBUG("KYTRACE SAVE_PCI_CONTROL_REGS");
    /* Disable interrupts before reprogramming. O/w the board will get into
     * a funny state and hang the system . */
    ACL_DEBUG (KERN_DEBUG "Saving PCI control registers");
    disable_aer_on_upstream_dev(aclpci);
//old    release_irq (aclpci->pci_dev, aclpci); 
//  release_irq (aclpci->pci_dev, aclpci,0);    
    result = pci_save_state(aclpci->pci_dev);
    break;
  }

  case ACLPCI_CMD_LOAD_PCI_CONTROL_REGS: {
    ACL_DEBUG("KYTRACE LOAD_PCI_CONTROL_REGS");
      
    pci_set_master(aclpci->pci_dev);
    init_irq (aclpci->pci_dev, aclpci);
    /*result = */pci_restore_state(aclpci->pci_dev);
    restore_aer_on_upstream_dev(aclpci);
    get_pcie_params(aclpci);
    ACL_DEBUG (KERN_DEBUG "Restored PCI control registers");
    break;
  }

  case ACLPCI_CMD_INTERRUPT_ENABLE: {
    ACL_DEBUG("KYTRACE INTERRUPT_ENABLE");
    result = init_irq(aclpci->pci_dev, aclpci);
    break;
  }

  case ACLPCI_CMD_INTERRUPT_DISABLE: {
    ACL_DEBUG("KYTRACE INTERRUPT_DISABLE");
    result = release_irq(aclpci->pci_dev, aclpci, 0);
    break;
  }

//  case ACLPCI_CMD_PIN_USER_ADDR:
//    result = aclpci_pin_user_addr (kcmd.user_addr, count);
 //   break;

//  case ACLPCI_CMD_UNPIN_USER_ADDR:
//    result = aclpci_unpin_user_addr (kcmd.user_addr, count);
 //   break;
    
 // case ACLPCI_CMD_GET_DMA_IDLE_STATUS: {
//    u32 idle = aclpci_dma_get_idle_status(aclpci);
//    result = copy_to_user ( kcmd.user_addr, &idle, sizeof(idle) );
 //   break;
//  }

//  case ACLPCI_CMD_DMA_UPDATE: {
//    aclpci_dma_update(aclpci, 0);
 //   break;
//  }

  case ACLPCI_CMD_GET_DEVICE_ID: {
    u32 id;
    ACL_DEBUG("KYTRACE GET_DEVICE_ID");
    id = aclpci->pci_dev->device;
    result = copy_to_user ( kcmd.user_addr, &id, sizeof(id) );
    break;
  }

  case ACLPCI_CMD_GET_VENDOR_ID: {
    u32 id;
    ACL_DEBUG("KYTRACE GET_VENDOR_ID");
    id = aclpci->pci_dev->vendor;
    result = copy_to_user ( kcmd.user_addr, &id, sizeof(id) );
    break;
  }

  case ACLPCI_CMD_GET_BUS: {
    u32 id;
    ACL_DEBUG("KYTRACE GET_BUS");
    id = aclpci->pci_dev->bus->number;
    result = copy_to_user ( kcmd.user_addr, &id, sizeof(id) );
    break;
  }

  case ACLPCI_CMD_GET_SLOT: {
    u32 id;
    ACL_DEBUG("KYTRACE GET_SLOT");
    id = PCI_SLOT(aclpci->pci_dev->devfn);
    result = copy_to_user ( kcmd.user_addr, &id, sizeof(id) );
    break;
  }

  case ACLPCI_CMD_GET_FUNC: {
    u32 id;
    ACL_DEBUG("KYTRACE GET_FUNC");
    id = PCI_FUNC(aclpci->pci_dev->devfn);
    result = copy_to_user ( kcmd.user_addr, &id, sizeof(id) );
    break;
  }

  case ACLPCI_CMD_GET_PCI_GEN: {
    u32 pci_gen;
    ACL_DEBUG("KYTRACE GET_PCI_GEN");
    pci_gen = aclpci->pci_gen;
    result = copy_to_user ( kcmd.user_addr, &pci_gen, sizeof(pci_gen) );
    break;
  }

  case ACLPCI_CMD_GET_PCI_NUM_LANES: {
    u32 pci_num_lanes;
    ACL_DEBUG("KYTRACE GET_PCI_NUM_LANES");
    pci_num_lanes = aclpci->pci_num_lanes;
    result = copy_to_user ( kcmd.user_addr, &pci_num_lanes, sizeof(pci_num_lanes) );
    break;
  }

  case ACLPCI_CMD_SET_SIGNAL_PAYLOAD: {
    u32 id;
    ACL_DEBUG("KYTRACE SET_SIGNAL_PAYLOAD");
    result = copy_from_user ( &id, kcmd.user_addr, sizeof(id) );
    aclpci->signal_info.si_int = id;
    break;
  }
  
  case ACLPCI_CMD_GET_DRIVER_VERSION: {
    ACL_DEBUG("KYTRACE GET_DRIVER_VERSION");
    /* Driver version is a string */
    result = copy_to_user ( kcmd.user_addr, &ACL_DRIVER_VERSION, strlen(ACL_DRIVER_VERSION)+1 );
    break;
  }

  //
  // KAYA Additional commands:
  //
  case ACLPCI_CMD_KAYA_DEBUG:
  {
      WD_DEBUG_INFO wdi;
      ACL_DEBUG("KYTRACE KAYA_DEBUG");
      get_wd_debug_info(&wdi);
//      int interrupts_counter = aclpci->num_handled_interrupts;
      if(0 != copy_to_user ( kcmd.user_addr, &wdi, sizeof(wdi)))
      {
          result = -EFAULT;
          break;
      }
      break;
  }
  
  case ACLPCI_CMD_DMA_SGLOCK: {
    //struct aclpci_dma *d = &(aclpci->dma_data);

    //void *addr = kcmd.user_addr;
    
/*@@@mbl commented  
#if (0 == DMA_LOCK_METHOD)
    int i;
    size_t firstPageLength = 0;
#endif
*/
    WD_DMA* pUserWdDMA = (WD_DMA*)kcmd.user_addr2;

    WD_DMA kernerWdDMA;
    WD_DMA_PAGE * pOriginalUserDMAPageArray;
    if(0 != copy_from_user(&kernerWdDMA, pUserWdDMA, sizeof(WD_DMA)))
    {
        ACL_DEBUG("KYTRACE SGLOCK error copy_from");    
        result = -EFAULT;
        break;
    }
    ACL_DEBUG("KYTRACE SGLOCK dwBytes %d dwPages %d pUserAddr 0x%lx",kernerWdDMA.dwBytes,kernerWdDMA.dwPages, (unsigned long)kernerWdDMA.pUserAddr);    

    
    // save the original dma page array address
    pOriginalUserDMAPageArray = kernerWdDMA.Page;

//why not allocating kernerWdDMA.Page here?    
    kernerWdDMA.Page=NULL;
    
    result = lock_dma_buffer(aclpci, &kernerWdDMA);
    
    if(result!=0)
    {
     ACL_DEBUG("KYTRACE SGLOCK lock_dma_buffer error %ld",result);    
     break;
    }
    // copy output of lock_dma_buffer() back to user space addressed by kcmd.user_addr2
    //ACL_DEBUG (KERN_WARNING " pages_rem==%d, first_page_offset=%d, last_page_offset=%d\n", pinned_mem_ptr->pages_rem, pinned_mem_ptr->first_page_offset, pinned_mem_ptr->last_page_offset);
   // ACL_DEBUG (KERN_WARNING " gma->len==%d, dma->num_pages=%d\n", pinned_mem_ptr->dma.len, pinned_mem_ptr->dma.num_pages);
/*@@@mbl commented  
#if (0 == DMA_LOCK_METHOD)
    for(i = 0; i < kernerWdDMA.dwPages; i++)
    {
        kernerWdDMA.Page[i].pPhysicalAddr = kernerWdDMA.dma_addrs[i];
        kernerWdDMA.Page[i].dwBytes = PAGE_SIZE;
    }

    firstPageLength = PAGE_SIZE - kernerWdDMA.first_page_offset;

    kernerWdDMA.Page[0].dwBytes = (kernerWdDMA.dwBytes <= firstPageLength) ? kernerWdDMA.dwBytes : firstPageLength;
    //ACL_DEBUG (KERN_WARNING " first page size =%d\n", wdDMA.Page[0].dwBytes);

    if( (kernerWdDMA.dwPages > 1)
         &&
         (kernerWdDMA.last_page_offset > 0))
    {
        kernerWdDMA.Page[kernerWdDMA.dwPages - 1].dwBytes = kernerWdDMA.last_page_offset;
    }
    //ACL_DEBUG (KERN_WARNING " last page size =%d\n", wdDMA.Page[dma.num_pages - 1].dwBytes);
#endif
*/
    // copy newly created WD_DMA_PAGE structure for each created page to user space
    if(0 != copy_to_user(pOriginalUserDMAPageArray, kernerWdDMA.Page, kernerWdDMA.dwPages * sizeof(WD_DMA_PAGE)))
    {
        ACL_DEBUG("KYTRACE SGLOCK error copy_to UserDMAPageArray");    
        result = -EFAULT;
    }
    
    if(kernerWdDMA.Page)
     kfree(kernerWdDMA.Page);

    // copy newly created WD_DMA structure to user space
    kernerWdDMA.Page = pOriginalUserDMAPageArray;
    if(0 != copy_to_user(pUserWdDMA, &kernerWdDMA, sizeof(WD_DMA)))
    {
        ACL_DEBUG("KYTRACE SGLOCK error copy_to pUserWdDMA");    
        result = -EFAULT;
    }
    break;
  }

#if (KAYA_CUDA_BUILD > 0)
  case ACLPCI_CMD_PIN_CUDA_DGPU: {
      ACL_DEBUG("KYTRACE");
      result = -48;
      break;
  }
#endif // #if (KAYA_CUDA_BUILD > 0)  

  case ACLPCI_CMD_DMA_SGUNLOCK: {

//    ACL_DEBUG("KYTRACE SGUNLOCK");    
    if(NULL != kcmd.user_addr2)
    {
        WD_DMA kernerWdDMA;
        if(0 != copy_from_user(&kernerWdDMA, kcmd.user_addr2, sizeof(WD_DMA)))
        {
            result = -EFAULT;
            break;
        }
        ACL_DEBUG("KYTRACE SGUNLOCK UserAddr 0x%lx",(unsigned long)kernerWdDMA.pUserAddr);    
        
        unlock_dma_buffer (aclpci, &kernerWdDMA);
    }
    else
     result = -EFAULT;
    break;
  }

  case ACLPCI_CMD_SYNC_FOR_CPU: {

    if(NULL != kcmd.user_addr2)
    {
        WD_DMA kernerWdDMA;

        if(0 != copy_from_user(&kernerWdDMA, kcmd.user_addr2, sizeof(WD_DMA)))
        {
            result = -EFAULT;
            break;
        }
        ACL_DEBUG("KYTRACE_SYNC_FOR_CPU  UserAddr 0x%lx",(unsigned long)kernerWdDMA.pUserAddr);

        sync_cache_for_memory(aclpci, &kernerWdDMA, false);

        //sync_dma_buffer (aclpci, &kernerWdDMA);
    }
    else
     result = -EFAULT;
      
    break;
  }

  case ACLPCI_CMD_SYNC_FOR_DEVICE: {
    
    if(NULL != kcmd.user_addr2)
    {
        WD_DMA kernerWdDMA;

        if(0 != copy_from_user(&kernerWdDMA, kcmd.user_addr2, sizeof(WD_DMA)))
        {
            result = -EFAULT;
            break;
        }
        ACL_DEBUG("KYTRACE_SYNC_FOR_DEVICE  UserAddr 0x%lx",(unsigned long)kernerWdDMA.pUserAddr);

        sync_cache_for_memory(aclpci, &kernerWdDMA, true);
    }
    else
     result = -EFAULT;
    break;
  }

  default:
    ACL_DEBUG (KERN_WARNING " Invalid command id %u! Ignoring the call. See aclpci_common.h for list of understood commands", kcmd.command);
    result = -EFAULT;
    break;
  }
  
//  ACL_DEBUG("KYTRACE %d",result);
  return result;
}




/* Pinning user pages.
 * 
 * Taken from <kernel code>/drivers/infiniband/hw/ipath/ipath_user_pages.c
 */
static void __aclpci_release_user_pages(struct page **p, size_t num_pages, int dirty)
{
    size_t i;

    for (i = 0; i < num_pages; i++) 
    {
        if (dirty) 
        {
            set_page_dirty_lock(p[i]);
        }
        put_page(p[i]);//page_cache_release(p[i]);//
    }
}

/* call with current->mm->mmap_sem held */
static int __aclpci_get_user_pages(unsigned long start_page, size_t num_pages,
                                   struct page **p, struct vm_area_struct **vma)
{
    size_t got;
    int ret;

//#define USE_get_user_pages_fast
#ifdef USE_get_user_pages_fast
    // see https://www.kernel.org/doc/htmldocs/kernel-api/API-get-user-pages-fast.html
    // for get_user_pages_fast documentation
    ACL_DEBUG (KERN_DEBUG "get_user_pages_fast %ld",num_pages);

    for (got = 0; got < num_pages; got += ret)
    {
        ret = get_user_pages_fast(
                             start_page + got * PAGE_SIZE,  // unsigned long start
                             num_pages - got,               // int nr_pages
                             1,                             // write
                             p + got                        // struct page ** pages
                             );
        if (ret < 0)
            goto bail_release;
    }
#else // #ifdef USE_get_user_pages_fast

    // see http://www.hep.by/gnu/kernel/kernel-api/API-get-user-pages.html
    // and http://www.makelinux.net/ldd3/chp-15-sect-3
    // for get_user_pages documentation

    for (got = 0; got < num_pages; got += ret)
    {
        
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,9,0)
   if(got==0)
    ACL_DEBUG (KERN_DEBUG "get_user_pages start_page FOLL_WRITE %ld",num_pages);
//  #pragma message "Building with 'gup_flags' parameter of 'get_user_pages'"        
       ret = get_user_pages(
        start_page + got * PAGE_SIZE,  // unsigned long start
        num_pages - got,               // int nr_pages
        FOLL_WRITE,                    // unsigned int gup_flags
        p + got,                       // struct page ** pages
        vma);                          // struct vm_area_struct ** vmas
#else
        int force = 0; // TODO: this was 1 originaly, changed to 0 following documentation (links above)

 #if LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0) //https://github.com/torvalds/linux/commit/d4edcf0d56958db0aca0196314ca38a5e730ea92
//  #pragma message "Building with 'write' and 'force' parameter of 'get_user_pages'"
   if(got==0)
     ACL_DEBUG (KERN_DEBUG "get_user_pages start_page %ld",num_pages);
       ret = get_user_pages(
        start_page + got * PAGE_SIZE,  // unsigned long start
        num_pages - got,               // int nr_pages
        1,                             // int write
        force,                         // int force
        p + got,                       // struct page ** pages
        vma);                          // struct vm_area_struct ** vmas
       
       
 #else      
//  #pragma message "Building with old 'get_user_pages' struct task_struct,... "
   if(got==0)
     ACL_DEBUG (KERN_DEBUG "get_user_pages current  %ld",num_pages);
       ret = get_user_pages(
        current,                       // struct task_struct * tsk
        current->mm,                   // struct mm_struct * mm
        start_page + got * PAGE_SIZE,  // unsigned long start
        num_pages - got,               // int nr_pages
        1,                             // int write
        force,                         // int force
        p + got,                       // struct page ** pages
        vma);                          // struct vm_area_struct ** vmas
 #endif        
#endif        
        if (ret < 0)
            goto bail_release;
    }
    // https://forum.blackmagicdesign.com/viewtopic.php?f=12&t=55075: #if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0) ...# elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 6, 0)
    // https://www.linuxquestions.org/questions/linux-kernel-70/direct-i-o-usgin-get_user_pages-4175559449/
    // https://lwn.net/Articles/753027/
#endif // else of #ifdef USE_get_user_pages_fast

    current->mm->locked_vm += num_pages;

    ret = 0;
    goto bail;

bail_release:
    //tickets #2334: we should use 'got' here
    __aclpci_release_user_pages(p, got, 0);
    ACL_DEBUG (KERN_DEBUG "'get_user_pages' failed with code %d at page %ld", ret,got);
bail:
    return ret;
}

/**
 * aclpci_get_user_pages - lock user pages into memory
 * @start_page: the start page
 * @num_pages: the number of pages
 * @p: the output page structures
 *
 * This function takes a given start page (page aligned user virtual
 * address) and pins it and the following specified number of pages.
 */
int aclpci_get_user_pages(unsigned long start_page, size_t num_pages, struct page **p)
{
    int ret;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)
    mmap_write_lock(current->mm);
#else
    down_write(&current->mm->mmap_sem);
#endif
    
    ret = __aclpci_get_user_pages(start_page, num_pages, p, NULL);
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)
    mmap_write_unlock(current->mm);
#else
    up_write(&current->mm->mmap_sem);
#endif

    return ret;
}

void aclpci_release_user_pages(struct page **p, size_t num_pages)
{
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)
    mmap_write_lock(current->mm);
#else
    down_write(&current->mm->mmap_sem);
#endif

    __aclpci_release_user_pages(p, num_pages, 1);

    current->mm->locked_vm -= num_pages;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)
    mmap_write_unlock(current->mm);
#else
    up_write(&current->mm->mmap_sem);
#endif
    
}


void get_pcie_params (struct aclpci_dev *aclpci)
{

  struct pci_dev *dev = aclpci->pci_dev;
  u16 linkstat, speed, width;
  struct pci_dev *upstream;
  int pos, upos;
  u16 status_reg, control_reg, link_cap_reg;
  u16 status, control;
  u32 link_cap;
  
  /* Defines for some special PCIe control bits */
  #define DISABLE_LINK_BIT         (1 << 4)
  #define RETRAIN_LINK_BIT         (1 << 5)
  #define TRAINING_IN_PROGRESS_BIT (1 << 11)
  #define LINKSPEED_2_5_GB         (0x1)
  #define LINKSPEED_5_0_GB         (0x2)
   
  pos = pci_find_capability (dev, PCI_CAP_ID_EXP);
  if (!pos) {
    ACL_DEBUG (KERN_WARNING "Can't find PCI Express capability!");
    return;
  }

  /* Find root node for this bus and tell it to retrain itself. */
  upstream = aclpci->upstream;
  if (upstream == NULL) {
    return;
  }
  upos = pci_find_capability (upstream, PCI_CAP_ID_EXP);
  status_reg = upos + PCI_EXP_LNKSTA;
  control_reg = upos + PCI_EXP_LNKCTL;
  link_cap_reg = upos + PCI_EXP_LNKCAP;
  pci_read_config_word (upstream, status_reg, &status);
  pci_read_config_word (upstream, control_reg, &control);
  pci_read_config_dword (upstream, link_cap_reg, &link_cap);
  
  
  pci_read_config_word (dev, pos + PCI_EXP_LNKSTA, &linkstat);
  pci_read_config_dword (upstream, link_cap_reg, &link_cap);
  speed = linkstat & PCI_EXP_LNKSTA_CLS;
  width = (linkstat & PCI_EXP_LNKSTA_NLW) >> PCI_EXP_LNKSTA_NLW_SHIFT;
  
  aclpci->pci_gen = speed;
  aclpci->pci_num_lanes = width;

  return;

}


/* For some reason, pci_find_ext_capability is not resolved
 * when loading this driver. So copied the implementation here. */
#define PCI_CFG_SPACE_SIZE 256
#define PCI_CFG_SPACE_EXP_SIZE 4096
 int my_pci_find_ext_capability(struct pci_dev *dev, int cap)
{
  u32 header;
  int ttl;
  int pos = PCI_CFG_SPACE_SIZE;

  /* minimum 8 bytes per capability */
  ttl = (PCI_CFG_SPACE_EXP_SIZE - PCI_CFG_SPACE_SIZE) / 8;

  if (dev->cfg_size <= PCI_CFG_SPACE_SIZE)
    return 0;

  if (pci_read_config_dword(dev, pos, &header) != PCIBIOS_SUCCESSFUL)
    return 0;

  /*
  * If we have no capabilities, this is indicated by cap ID,
  * cap version and next pointer all being 0.
  */
  if (header == 0)
    return 0;

  while (ttl-- > 0) {
  if (PCI_EXT_CAP_ID(header) == cap)
    return pos;

  pos = PCI_EXT_CAP_NEXT(header);
  if (pos < PCI_CFG_SPACE_SIZE)
    break;

  if (pci_read_config_dword(dev, pos, &header) != PCIBIOS_SUCCESSFUL)
    break;
  }

  return 0;
}


/* return value of AER uncorrectable error mask register
 * for UPSTREAM node of the given device. */
u32 get_aer_uerr_mask_reg (struct aclpci_dev *aclpci)
{
  struct pci_dev *dev = aclpci->upstream;
  u32 reg32 = 0;
  int pos;

  if (dev == NULL) {
    ACL_DEBUG (KERN_DEBUG "No upstream device found!");
    return -EIO;
  }
  
  //TODO: compilation fails on RedHat
//  if (dev->__aer_firmware_first) {
//    return -EIO;
//  }

  pos = my_pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
  if (!pos) {
    ACL_DEBUG (KERN_DEBUG "Upstream device doesn't have AER extended capability.");
    return -EIO;
  }
  
  pci_read_config_dword(dev, pos+0x4, &reg32);
  pci_read_config_dword(dev, pos+0x8, &reg32);
  return reg32;
}

/* Surprise down is the 5th register inside AER uncorrectable error register/mask */
#define AER_SURPRISE_DOWN 0x20


/* Set AER uncorrectable error mask register
 * for UPSTREAM node of the given device. */
void set_aer_uerr_mask_reg (struct aclpci_dev *aclpci, u32 val)
{
  int pos;
  struct pci_dev *dev = aclpci->upstream;
  if (!dev) {
    return;
  }
    
  pos = my_pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
  if (!pos) {
    return;
  }

  /* First, clear the error bit by writing 1 to 5th bit. */
  pci_write_config_dword(dev, pos+0x4, AER_SURPRISE_DOWN);

  /* Now set the mask register */
  pci_write_config_dword(dev, pos+0x8, val);
  return;
}


/* Mask off "Surprise Down" error in AER. Note that setting the mask to '1' means
 * the error is ignored. */
void disable_aer_on_upstream_dev(struct aclpci_dev *aclpci) {

  u32 disabled;
  aclpci->aer_uerr_mask_reg = get_aer_uerr_mask_reg(aclpci);
  if (aclpci->aer_uerr_mask_reg == -EIO) {
    return;
  }
    
  disabled = aclpci->aer_uerr_mask_reg | AER_SURPRISE_DOWN;
  
  ACL_DEBUG (KERN_WARNING "Changing AER Uncorrectable error mask register from %x to %x", 
             aclpci->aer_uerr_mask_reg, disabled);
  set_aer_uerr_mask_reg(aclpci, disabled);
}


/* Restore AER uncorrectable error mask register
 * for UPSTREAM node of the given device. */
void restore_aer_on_upstream_dev(struct aclpci_dev *aclpci) {

  if (aclpci->aer_uerr_mask_reg == -EIO)
    return;
  ACL_DEBUG (KERN_WARNING "Restoring AER Uncorrectable error mask register to %x", aclpci->aer_uerr_mask_reg);
  set_aer_uerr_mask_reg(aclpci, aclpci->aer_uerr_mask_reg);
}
